@@ -1,5 +1,5 @@
 /*
- Copyright (C) 2010, Philipp Merkel <linux@philmerk.de>
+ Copyright (C) 2010, 2011 Philipp Merkel <linux@philmerk.de>
 
  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
@@ -33,7 +33,7 @@ int clickMode;
 
 /* Number of milliseconds before a single click is registered, to give the user time to put down
    second finger for two-finger gestures. */
-#define CLICK_DELAY 200
+#define CLICK_DELAY 100
 
 /* Continuation mode -- when 1, two finger gesture is continued when one finger is released. */
 #define CONTINUATION 1
@@ -52,7 +52,7 @@ int clickMode;
 /* Maximum distance that two fingers have been moved while they were on */
 double maxDist;
 /* The time when the first finger touched */
-Time fingerDownTime;
+TimeVal fingerDownTime;
 /* Were there once two fingers on during the current touch phase? */
 int hadTwoFingersOn = 0;
 
@@ -80,8 +80,8 @@ int lastScrollDirectionX = 0;
 /* 1 if the last Y scroll command was down, -1 if it was up. */
 int lastScrollDirectionY = 0;
 /* Time last scroll command in X/Y action was sent. */
-Time lastScrollXTime;
-Time lastScrollYTime;
+TimeVal lastScrollXTime;
+TimeVal lastScrollYTime;
 /* Interval between the last two X/Y scroll commands. */
 int lastScrollXIntv;
 int lastScrollYIntv;
@@ -116,6 +116,9 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 		maxDist = moveDist;
 	}
 
+	/* Prevent division by zero */
+	if(gestureStartDist < 1) gestureStartDist = 0;
+
 	/* We don't know yet what to do, so look if we can decide now (only do this if there
 	   are still two fingers down, otherwise we are in continuation and can't decide). */
 	if (amPerformingGesture == GESTURE_UNDECIDED && fingersDown == 2) {
@@ -149,9 +152,12 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 		}
 
 		int zoomMinDist = currentProfile->zoomMinDistance;
-		if (currentProfile->zoomInherit)
+		double zoomMinFactor = currentProfile->zoomMinFactor;
+		if (currentProfile->zoomInherit) {
 			zoomMinDist = defaultProfile.zoomMinDistance;
-		if (abs((int) currentDist - gestureStartDist) > zoomMinDist) {
+			zoomMinFactor = defaultProfile.zoomMinFactor;
+		}
+		if (abs((int) currentDist - gestureStartDist) > zoomMinDist && (currentDist / gestureStartDist > zoomMinFactor || currentDist / gestureStartDist < 1/zoomMinFactor)) {
 			amPerformingGesture = GESTURE_ZOOM;
 			if(inDebugMode()) printf("Start zoom gesture\n");
 			return 1;
@@ -177,7 +183,7 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 		}
 	}
 
-	/* If we know hat gesture to perform, look if there is something to do */
+	/* If we know what gesture to perform, look if there is something to do */
 	switch (amPerformingGesture) {
 	case GESTURE_SCROLL:
 		;
@@ -192,11 +198,11 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 		if (hscrollStep == 0 || vscrollStep == 0)
 			return 0;
 
-		Time currentTime = getCurrentTime();
+		TimeVal currentTime = getCurrentTime();
 		if (hscrolledBy > hscrollStep) {
 			lastScrollDirectionX = 1;
 			lastLastScrollXIntv = lastScrollXIntv;
-			lastScrollXIntv = currentTime - lastScrollXTime;
+			lastScrollXIntv = timeDiff(lastScrollXTime, currentTime);
 			lastScrollXTime = currentTime;
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollRightAction),
@@ -210,7 +216,7 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 			return 1;
 		} else if (hscrolledBy < -hscrollStep) {
 			lastLastScrollXIntv = lastScrollXIntv;
-			lastScrollXIntv = currentTime - lastScrollXTime;
+			lastScrollXIntv = timeDiff(lastScrollXTime, currentTime);
 			lastScrollXTime = currentTime;
 			lastScrollDirectionX = -1;
 			if (currentProfile->scrollInherit) {
@@ -226,7 +232,7 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 		}
 		if (vscrolledBy > vscrollStep) {
 			lastLastScrollYIntv = lastScrollYIntv;
-			lastScrollYIntv = currentTime - lastScrollYTime;
+			lastScrollYIntv = timeDiff(lastScrollYTime, currentTime);
 			lastScrollYTime = currentTime;
 			lastScrollDirectionY = 1;
 			if (currentProfile->scrollInherit) {
@@ -241,7 +247,7 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 			return 1;
 		} else if (vscrolledBy < -vscrollStep) {
 			lastLastScrollYIntv = lastScrollYIntv;
-			lastScrollYIntv = currentTime - lastScrollYTime;
+			lastScrollYIntv = timeDiff(lastScrollYTime, currentTime);
 			lastScrollYTime = currentTime;
 			lastScrollDirectionY = -1;
 			if (currentProfile->scrollInherit) {
@@ -333,10 +339,10 @@ int checkGesture(FingerInfo* fingerInfos, int fingersDown) {
 void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersWereDown) {
 
 	if(fingersDown != 0 && fingersWereDown == 0) {
-		stopEasingThread();
+		stopEasing();
 	}
 
-	Time currentTime = getCurrentTime();
+	TimeVal currentTime = getCurrentTime();
 	if (TWO_FINGERS_DOWN) {
 		/* Second finger touched (and maybe first too) */
 
@@ -351,7 +357,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		hadTwoFingersOn = 1;
 
 		/* Get current profile */
-		currentProfile = getWindowProfile(getCurrentWindow());
+		currentProfile = getWindowProfile(getActiveWindow());
 		if(inDebugMode()) {
 			if(currentProfile->windowClass != NULL) {
 				printf("Use profile '%s'\n", currentProfile->windowClass);
@@ -360,7 +366,6 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 			}
 		}
 
-
 		/* If there had already been a single-touch event raised because the
 		 * user was too slow, stop it now. */
 		releaseButton();
@@ -377,7 +382,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		/* We have not decided on a gesture yet. */
 		amPerformingGesture = GESTURE_UNDECIDED;
 
-		movePointer(gestureStartCenterX, gestureStartCenterY);
+		movePointer(gestureStartCenterX, gestureStartCenterY, 0);
 	} else if (TWO_FINGERS_ON) {
 
 		/* Moved with two fingers */
@@ -389,7 +394,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		} else {
 			int i;
 			for(i = 0; i <= 1; i++) {
-				if(fingerInfos[i].id != -1) {
+				if(fingerInfos[i].slotUsed) {
 					currentCenterX = fingerInfos[i].x;
 					currentCenterY = fingerInfos[i].y;
 				}
@@ -400,7 +405,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		 * we need to move the pointer */
 		if (amPerformingGesture == GESTURE_SCROLL && dragScrolling) {
 			/* Move pointer to center between touch points */
-			movePointer(currentCenterX, currentCenterY);
+			movePointer(currentCenterX, currentCenterY, 0);
 		}
 
 		/* Perform gestures as long as there are some. */
@@ -431,8 +436,8 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 
 				/* Check if scrolling intervals are not too long. Also check if last scrolling on an axis is longer
 				   ago than twice its interval, which means the scrolling has been stopped or extremely slowed down since. */
-				if(lastScrollYIntv == 0 || currentTime - lastScrollYTime > lastScrollYIntv * 2 || lastScrollYIntv > MAX_EASING_START_INTERVAL) dirY = 0;
-				if(lastScrollXIntv == 0 || currentTime - lastScrollXTime > lastScrollXIntv * 2 || lastScrollXIntv > MAX_EASING_START_INTERVAL) dirX = 0;
+				if(lastScrollYIntv == 0 || timeDiff(lastScrollYTime, currentTime) > lastScrollYIntv * 2 || lastScrollYIntv > MAX_EASING_START_INTERVAL) dirY = 0;
+				if(lastScrollXIntv == 0 || timeDiff(lastScrollXTime, currentTime) > lastScrollXIntv * 2 || lastScrollXIntv > MAX_EASING_START_INTERVAL) dirX = 0;
 				if(dirX != 0 || dirY != 0) {
 					if(dirX != 0 && dirY != 0) {
 						/* As we only support one interval, only use larger axis. */
@@ -462,10 +467,11 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 				== GESTURE_UNDECIDED) && maxDist < 10) {
 			/* Move pointer to correct position */
 			if(clickMode == 2) {
-				movePointer(gestureStartCenterX, gestureStartCenterY);
+				/* Assume first finger is at ID 0 and second finger at ID 1, might have to be changed later */
+				movePointer(gestureStartCenterX, gestureStartCenterY, fingerInfos[0].rawZ);
 			} else {
 				/* Assume first finger is at ID 0 and second finger at ID 1, might have to be changed later */
-				movePointer(fingerInfos[clickMode].x, fingerInfos[clickMode].y);
+				movePointer(fingerInfos[clickMode].x, fingerInfos[clickMode].y, fingerInfos[clickMode].rawZ);
 			}
 
 			if (currentProfile->tapInherit) {
@@ -483,15 +489,15 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		/* Fake single-touch move event */
 		int i;
 		for(i = 0; i <= 1; i++) {
-			if(fingerInfos[i].id != -1) {
-				movePointer(fingerInfos[i].x, fingerInfos[i].y);
+			if(fingerInfos[i].slotUsed) {
+				movePointer(fingerInfos[i].x, fingerInfos[i].y, fingerInfos[i].rawZ);
 			}
 		}
 
 	} else if (fingersDown == 1) {
 		/* Moved with one finger */
 		if (hadTwoFingersOn == 0 && !isButtonDown()) {
-			if (currentTime > fingerDownTime + CLICK_DELAY) {
+			if (timeDiff(fingerDownTime, currentTime) > CLICK_DELAY) {
 				/* Delay has passed, no gesture been performed, so perform single-touch press now */
 				pressButton();
 			}
@@ -500,13 +506,14 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 			/* Fake single-touch move event */
 			int i;
 			for(i = 0; i <= 1; i++) {
-				if(fingerInfos[i].id != -1) {
-					movePointer(fingerInfos[i].x, fingerInfos[i].y);
+				if(fingerInfos[i].slotUsed) {
+					movePointer(fingerInfos[i].x, fingerInfos[i].y, fingerInfos[i].rawZ);
 				}
 			}
 
 		}
 	} else if (fingersDown == 0 && fingersWereDown > 0) {
+		printf("Last finger released!\n");
 		/* Last finger released */
 		if (hadTwoFingersOn == 0 && !isButtonDown()) {
 			/* The button press time has not been reached yet, and we never had two
@@ -514,9 +521,11 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 			 * we simulate button down and up now. */
 			pressButton();
 			releaseButton();
+			printf("Handle release 1\n");
 		} else {
 			/* We release the button if it is down. */
 			releaseButton();
+			printf("Handle release 2\n");
 		}
 	}
 	if (fingersDown == 0) {
@@ -592,7 +601,7 @@ int isWindowBlacklistedForGestures(Window w) {
 	} else {
 		if(inDebugMode()) printf("Found window with id %i and no class.\n", (int) w);
 		//if(inDebugMode()) printf("Look for another child\n");
-		//return isWindowBlacklisted(getLastChildWindow(w));
+		return isWindowBlacklisted(getLastChildWindow(w));
 		return 0;
 	}
 }
