@@ -24,7 +24,7 @@
 char* blacklist[] = { "xournal", "gimp-2.6", "gimp", "mypaint", "inkscape", NULL };
 
 /* WM-Blacklist */
-char* wmBlacklist[] = { "kwin", "mutter", NULL };
+char* wmBlacklist[] = { "kwin", "mutter", "metacity", "compiz", "unity", NULL };
 
 /* Profiles */
 int profileCount = 8;
@@ -124,8 +124,8 @@ Profile profiles[] = {
 					.hscrollStep = 300,
 					.vscrollStep = 125,
 					.scrollBraceAction = { ACTIONTYPE_NONE, 0, 0 },
-					.scrollUpAction = { ACTIONTYPE_BUTTONPRESS,5,0 },
-					.scrollDownAction = { ACTIONTYPE_BUTTONPRESS,4,0 },
+					.scrollUpAction = { ACTIONTYPE_KEYPRESS,XK_Down, MODIFIER_CONTROL | MODIFIER_ALT },
+					.scrollDownAction = { ACTIONTYPE_KEYPRESS,XK_Up, MODIFIER_CONTROL | MODIFIER_ALT },
 					.scrollLeftAction = { ACTIONTYPE_KEYPRESS,XK_Right, MODIFIER_CONTROL | MODIFIER_ALT },
 					.scrollRightAction = { ACTIONTYPE_KEYPRESS,XK_Left, MODIFIER_CONTROL | MODIFIER_ALT },
 					.scrollEasing = 0,
