@@ -1,5 +1,5 @@
 /*
- Copyright (C) 2010, Philipp Merkel <linux@philmerk.de>
+ Copyright (C) 2010, 2011 Philipp Merkel <linux@philmerk.de>
 
  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
@@ -18,8 +18,8 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
-#include <pthread.h>
 #include <X11/Xutil.h>
+#include <X11/X.h>
 #include <X11/Xos.h>
 #include <X11/Xatom.h>
 #include <X11/extensions/XTest.h>
@@ -28,8 +28,11 @@
 #include <X11/extensions/Xrandr.h>
 #include "twofingemu.h"
 #include "gestures.h"
+#include "easing.h"
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/select.h>
+#include <time.h>
 
 #define EXIT_SUCCESS 0
 #define EXIT_FAILURE 1
@@ -84,8 +87,8 @@ static void daemonize(void) {
 }
 
 /* Finger information */
-FingerInfo fingerInfos[2] = { { .rawX=0, .rawY=0, .id = -1, .setThisTime = 0 }, { .rawX=0, .rawY=0,
-		.id = -1, .setThisTime = 0 } };
+FingerInfo fingerInfos[2] = { { .rawX=0, .rawY=0, .rawZ=0, .id = -1, .slotUsed = 0, .setThisTime = 0 }, { .rawX=0, .rawY=0, .rawZ=0,
+		.id = -1, .slotUsed = 0, .setThisTime = 0 } };
 
 /* X stuff */
 Display* display;
@@ -94,8 +97,6 @@ int screenNum;
 int deviceID;
 Atom WM_CLASS;
 pthread_t xLoopThread;
-Time fairlyCurrentTime;
-
 
 /* Calibration data */
 int calibMinX, calibMaxX, calibMinY, calibMaxY;
@@ -106,18 +107,14 @@ int calibMatrixUse = 0;
 /* The width and height of the screen in pixels */
 unsigned int screenWidth, screenHeight;
 
-/* Two finger gestures activated (and input grabbed)? */
-int active = 0;
-/* Activate two finger gestures (and grab input) as soon as the fingers have been removed */
-int activateAtRelease = 0;
-
 /* Finger data */
 int fingersDown = 0;
 int fingersWereDown = 0;
 /* Has button press of first button been called in XTest? */
 int buttonDown = 0;
 
-
+/* Does the device use the legacy MT protocol? */
+int useLegacyProtocol = 0;
 
 /* Handle errors by, well, throwing them away. */
 int invalidWindowHandler(Display *dsp, XErrorEvent *err) {
@@ -127,62 +124,87 @@ int invalidWindowHandler(Display *dsp, XErrorEvent *err) {
 /* Grab the device so input is captured */
 void grab(Display* display, int grabDeviceID) {
 	XIEventMask device_mask;
-	unsigned char mask_data[1] = { 0 };
+	unsigned char mask_data[8] = { 0,0,0,0,0,0,0,0 };
 	device_mask.mask_len = sizeof(mask_data);
 	device_mask.mask = mask_data;
-	XISetMask(device_mask.mask, XI_ButtonPress);
-	XISetMask(device_mask.mask, XI_ButtonRelease);
+	//XISetMask(device_mask.mask, XI_TouchBegin);
+	//XISetMask(device_mask.mask, XI_TouchUpdate);
+	//XISetMask(device_mask.mask, XI_TouchEnd);
 	XISetMask(device_mask.mask, XI_Motion);
+	XISetMask(device_mask.mask, XI_ButtonPress);
 
-	XIGrabModifiers modifiers[1] = { { 0, 0 } };
+//	XIGrabModifiers modifiers;
+//	modifiers.modifiers = XIAnyModifier;
+//	int r = XIGrabButton(display, grabDeviceID, XIAnyButton, root, None, GrabModeSync,
+//			GrabModeAsync, True, &device_mask, 1, &modifiers);
 
-	XIGrabButton(display, grabDeviceID, 1, root, None, GrabModeAsync,
-			GrabModeAsync, False, &device_mask, 1, modifiers);
+//	int r = XIGrabTouchBegin(display, grabDeviceID, root, None, &device_mask, 0, modifiers);
+
+	int r = XIGrabDevice(display, grabDeviceID, root, CurrentTime, None, GrabModeAsync, GrabModeAsync, False, &device_mask);
+
+	if(debugMode) printf("Grab Result: %i\n", r);
 
 }
 
 /* Ungrab the device so input can be handled by application directly */
 void ungrab(Display* display, int grabDeviceID) {
-	XIGrabModifiers modifiers[1] = { { 0, 0 } };
-	XIUngrabButton(display, grabDeviceID, 1, root, 1, modifiers);
-
+//	XIGrabModifiers modifiers[1] = { { 0, 0 } };
+//	XIUngrabButton(display, grabDeviceID, 1, root, 1, modifiers);
+	XIUngrabDevice(display, grabDeviceID, CurrentTime);
 }
 
-Time getCurrentTime() {
-	return fairlyCurrentTime;
+TimeVal getCurrentTime() {
+
+	TimeVal time;
+	gettimeofday(&time, NULL);
+
+	return time;
+}
+
+int timeDiff(TimeVal start, TimeVal end)
+{
+	long seconds  = end.tv_sec  - start.tv_sec;
+	long microSeconds = end.tv_usec - start.tv_usec;
+
+	return (seconds * 1000 + microSeconds/1000);
 }
 
 
 
-/* Activates two finger gesture handling and input grabbing */
-void activate() {
-	activateAtRelease = 0;
-	if (active == 0) {
-		grab(display, deviceID);
-		active = 1;
-	}
-}
-
-/* Deactivates two finger gesture handling and input grabbing */
-void deactivate() {
-	activateAtRelease = 0;
-	if (active) {
-		active = 0;
-		ungrab(display, deviceID);
-	}
-}
-
 /* Send an XTest event to release the first button if it is currently pressed */
 void releaseButton() {
 	if (buttonDown) {
 		buttonDown = 0;
 		XTestFakeButtonEvent(display, 1, False, CurrentTime);
+		
+		/*printf("RELEASE!\n");
+		XDevice * dev = XOpenDevice(display, deviceID);
+		int axes[3] = {fingerInfos[0].rawX, fingerInfos[0].rawY, fingerInfos[0].rawZ};
+		XTestFakeDeviceButtonEvent(display, dev, 1, False, axes, 3, 0);
+		XCloseDevice(display, dev);
+
 		XFlush(display);
+		grab(display, deviceID);*/
 	}
 }
 /* Send an XTest event to press the first button if it is not pressed yet */
 void pressButton() {
 	if(!buttonDown) {
+
+/*		XDevice * dev = XOpenDevice(display, deviceID);
+		int axes[3] = {fingerInfos[0].rawX, fingerInfos[0].rawY, fingerInfos[0].rawZ};
+		XTestFakeDeviceButtonEvent(display, dev, 1, False, axes, 3, 0);
+		XCloseDevice(display, dev);
+		XFlush(display);
+
+		ungrab(display, deviceID);
+		XFlush(display);
+
+		printf("PRESS!\n");
+		dev = XOpenDevice(display, deviceID);
+		XTestFakeDeviceButtonEvent(display, dev, 1, True, axes, 3, 0);
+		XCloseDevice(display, dev);
+*/
 		buttonDown = 1;
 		XTestFakeButtonEvent(display, 1, True, CurrentTime);
 		XFlush(display);
@@ -195,8 +217,16 @@ int isButtonDown() {
 
 
 /* Moves the pointer to the given position */
-void movePointer(int x, int y) {
+void movePointer(int x, int y, int z) {
 	/* Move pointer to center between touch points */
+	/* TODO: Find some way to send XI events to also send pressure level */
+//	printf("%i %i %i\n", x, y, z);
+//	int axes[3] = {x, y, z};
+	
+//	XDevice * dev = XOpenDevice(display, 17);
+//	XTestFakeDeviceMotionEvent(display, dev, False, 0, axes, 2, 0);
+//	XCloseDevice(display, dev);
+
 	XTestFakeMotionEvent(display, -1, x, y, CurrentTime);
 	XFlush(display);
 }
@@ -310,7 +340,7 @@ Window getLastChildWindow(Window w) {
 
 		if (childWindows != NULL) {
 			if(childCount > 0) {
-				printf("%i children.\n", childCount);
+				//printf("%i children.\n", childCount);
 				Window child = childWindows[childCount - 1];
 				XFree(childWindows);
 				return child;
@@ -324,10 +354,26 @@ Window getLastChildWindow(Window w) {
 
 }
 
+Window getActiveWindow() {
+
+/*	Window *root_return = None, *child_return = None;
+	int *root_x_return = 0, *root_y_return = 0;
+	int *win_x_return = 0, *win_y_return = 0;
+	unsigned int *mask_return = 0;
+
+	XQueryPointer(display, root, root_return, child_return, root_x_return, root_y_return, 
+                     win_x_return, win_y_return, mask_return);
+
+	return *child_return;*/
+	return getCurrentWindow();
+
+}
+
 /* Returns the active top-level window. A top-level window is one that has WM_CLASS set.
  * May also return None. */
 Window getCurrentWindow() {
 	//if(debugMode) printf("Called getCurrentWindow\n");
+
 	/* First get the window that has the input focus */
 	Window currentWindow;
 	int revert;
@@ -434,47 +480,16 @@ void processFingers() {
 	int i;
 	fingersDown = 0;
 	for(i = 0; i < 2; i++) {
-		if(fingerInfos[i].id != -1) {
+		if(fingerInfos[i].slotUsed) {
 			calibrate(&(fingerInfos[i]));
 			fingersDown++;
 		}
 	}
 
-	/* If we should activate at release, do it. */
-	if (buttonDown == 0 && fingersDown == 0 && activateAtRelease != 0) {
-		releaseButton();
-		activate();
-	}
-
-	if (active == 0)
-		return;
-
 	processFingerGesture(fingerInfos, fingersDown, fingersWereDown);
 
 	/* Save number of fingers to compare next time */
 	fingersWereDown = fingersDown;
-
-	/* If we should activate at release, do it (now again because fingersDown might have changed). */
-	if (buttonDown == 0 && fingersDown == 0 && activateAtRelease != 0) {
-		releaseButton();
-		activate();
-	}
-}
-
-/* Called when a blacklisted window is left (for some reasons sometimes also if it's not blacklisted) */
-void leaveWindow() {
-	if (active == 0) {
-
-		activateAtRelease = 1;
-	}
-}
-
-/* Called when a blacklisted window is focused */
-void enterBlacklistedWindow() {
-	//printf("Entered bad window.\n");
-	activateAtRelease = 0;
-	releaseButton();
-	deactivate();
 }
 
 /* Returns a pointer to the profile of the currently selected
@@ -515,30 +530,6 @@ int isWindowBlacklisted(Window w) {
 	return isWindowBlacklistedForGestures(w);
 }
 
-/* Called when a new window is mapped. Checks whether it is blacklisted and registers
- * the activation events if yes. Then checks if it is active and, if yes, calls enter/leave.
- */
-void windowMapped(Window w) {
-
-	if (isWindowBlacklisted(w)) {
-		if(debugMode) printf("It's blacklisted.\n");
-		/* register for window focus events */
-		XSelectInput(display, w, EnterWindowMask | LeaveWindowMask);
-		if (getCurrentWindow() == w) {
-			/* If this is current and blacklisted, call enterBlacklistedWindow */
-			enterBlacklistedWindow();
-		}
-	} else {
-		if(debugMode) printf("It's not blacklisted.\n");
-		if (getCurrentWindow() == w) {
-			if(debugMode) printf("It's the current one!\n");
-			/* It is current and not blacklisted, so we might have left a blacklisted window */
-			leaveWindow();
-		}
-	}
-
-}
-
 void setScreenSize(XRRScreenChangeNotifyEvent * evt) {
 	screenWidth = evt->width;
 	screenHeight = evt->height;
@@ -550,66 +541,94 @@ void setScreenSize(XRRScreenChangeNotifyEvent * evt) {
 
 
 /* X thread that processes X events in parallel to kernel device loop */
-void* xLoopThreadFunction(void* arg) {
+void handleXEvent() {
 	XEvent ev;
-	while (1) {
-		XNextEvent(display, &ev);
+	XNextEvent(display, &ev);
+	//if(debugMode) printf("Handle event\n");
+	if (XGetEventData(display, &(ev.xcookie))) {
+		XGenericEventCookie *cookie = &(ev.xcookie);
 
-		if (XGetEventData(display, &(ev.xcookie))) {
-			XGenericEventCookie *cookie = &(ev.xcookie);
-			if (cookie->evtype == XI_Motion) {
-				/* Move event */
-				XIDeviceEvent* data = cookie->data;
-				fairlyCurrentTime = data->time;
-//				if (!dontMove) {
-//					/* Fake single-touch move event */
-//					XTestFakeMotionEvent(display, -1, (int) data->root_x,
-//							(int) data->root_y, CurrentTime);
-//					XFlush(display);
-//				}
-			} else if (cookie->evtype == XI_PropertyEvent) {
-				/* Device properties changed -> recalibrate. */
-				printf("Device properties changed.\n");
-				readCalibrationData(0);
-			}
-
-			XFreeEventData(display, &(ev.xcookie));
-
-		} else {
-			if (ev.type == MapNotify) {
-				windowMapped(ev.xmap.window);
-			} else if (ev.type == EnterNotify) {
-				/* This is only called for blacklisted windows */
-				enterBlacklistedWindow();
-			} else if (ev.type == LeaveNotify) {
-				/* This should only be called for blacklisted windows, but it's sometimes also
-				 * called when a non-blacklisted window is left. But that's no problem for us. */
-				leaveWindow();
-			} else if(ev.type == 101) {
-				/* Why isn't this magic constant explained anywhere?? */
-				setScreenSize((XRRScreenChangeNotifyEvent *) &ev);
-			}
+		// Touch events don't work right now
+		//if (cookie->evtype == XI_TouchBegin) {
+		//	if(debugMode) printf("Touch begin\n");
+		//} else if (cookie->evtype == XI_TouchUpdate) {
+		//	if(debugMode) printf("Touch update\n");
+		//} else if (cookie->evtype == XI_TouchEnd) {
+		//	if(debugMode) printf("Touch end\n");
+		if (cookie->evtype == XI_Motion) {
+			if(debugMode) printf("Motion event\n");
+			//int r = XIAllowEvents(display, deviceID, XIReplayDevice, CurrentTime);
+			//printf("XIAllowEvents result: %i\n", r);
+		} else if (cookie->evtype == XI_PropertyEvent) {
+			/* Device properties changed -> recalibrate. */
+			if(debugMode) printf("Device properties changed.\n");
+			readCalibrationData(0);
+		} else if(cookie->evtype == XI_ButtonPress) {
+			if(debugMode) printf("Button Press\n");
+			//int r = XIAllowEvents(display, deviceID, XIReplayDevice, CurrentTime);
+			//printf("XIAllowEvents result: %i\n", r);
 		}
-	}
-}
 
-/* Checks for all running windows if they are blacklisted and/or the current window.
- * Registers enter/leave event handlers for blacklisted windows. Executes them for
- * the current window. */
-void checkRunningWindows() {
-	Window aroot;
-	Window parent;
-	Window* childWindows = NULL;
-	unsigned int childCount;
-	/* Get all top-level windows */
-	if(XQueryTree(display, root, &aroot, &parent, &childWindows, &childCount)) {
 
-		if (childWindows != NULL) {
-			int i;
-			for (i = 0; i < childCount; i++) {
-				windowMapped(childWindows[i]);
-			}
-			XFree(childWindows);
+		// In an ideal world, the following would work. But unfortunately the touch events
+		// delivered by evdev are often crap on the eGalax screen, with missing events when there
+		// should be some. So we still have to read directly from the input device, as bad as that is.
+		//if (cookie->evtype == XI_TouchBegin || cookie->evtype == XI_TouchUpdate || cookie->evtype == XI_TouchEnd) {
+		//	XIDeviceEvent * devEvt = (XIDeviceEvent*) cookie->data;
+		//	printf("Detail: %i\n", devEvt->detail);
+		//	printf("Mask[0]: %i\n", (int) devEvt->valuators.mask[0]);
+		//
+		//	/* Look for slot to put the data into by looking at the tracking ids */
+		//	int index = -1;
+		//	int i;
+		//	for(i = 0; i < 2; i++) {
+		//		if(fingerInfos[i].id == devEvt->detail) {
+		//			index = i;
+		//			break;
+		//		}
+		//	}
+		//
+		//	/* No slot for this id found, look for free one */
+		//	if(index == -1) {
+		//		for(i = 0; i < 2; i++) {
+		//			if(!fingerInfos[i].slotUsed) {
+		//				/* "Empty" slot, so we can add it. */
+		//				index = i;
+		//				fingerInfos[i].id = devEvt->detail;
+		//				break;
+		//			}
+		//		}
+		//	}
+		//
+		//	/* We have found a slot */
+		//	if(index != -1) {
+		//		fingerInfos[index].slotUsed = (cookie->evtype != XI_TouchEnd ? 1 : 0);
+		//
+		//		i = 0;
+		//		if((devEvt->valuators.mask[0] & 1) == 1)
+		//		{
+		//			fingerInfos[index].rawX = (int) devEvt->valuators.values[i++];
+		//		}
+		//		if((devEvt->valuators.mask[0] & 2) == 2)
+		//		{
+		//			fingerInfos[index].rawY = (int) devEvt->valuators.values[i++];
+		//		}
+		//		if((devEvt->valuators.mask[0] & 4) == 4)
+		//		{
+		//			fingerInfos[index].rawZ = (int) devEvt->valuators.values[i++];
+		//		}
+		//	}
+		//
+		//	processFingers();
+		//}
+
+
+		XFreeEventData(display, &(ev.xcookie));
+
+	} else {
+		if(ev.type == 101) {
+			/* Why isn't this magic constant explained anywhere?? */
+			setScreenSize((XRRScreenChangeNotifyEvent *) &ev);
 		}
 	}
 }
@@ -795,7 +814,7 @@ int main(int argc, char **argv) {
 	}
 
 	/* We have two threads accessing X */
-	XInitThreads();
+	//XInitThreads();
 
 	/* Connect to X server */
 	if ((display = XOpenDisplay(NULL)) == NULL) {
@@ -823,8 +842,8 @@ int main(int argc, char **argv) {
 		devname = "/dev/twofingtouch";
 	}
 
-	/* Save return value of pthread_create so we only call it later if it hasn't been successfully called before. */
-	int threadReturn = 1;
+	///* Save return value of pthread_create so we only call it later if it hasn't been successfully called before. */
+	//int threadReturn = 1;
 
 	/* Try to read from device file */
 	int fileDesc;
@@ -833,6 +852,11 @@ int main(int argc, char **argv) {
 		return 1;
 	}
 
+	fd_set fileDescSet;
+	FD_ZERO(&fileDescSet);
+
+	int eventQueueDesc = XConnectionNumber(display);	
+
 	while (1) {
 		/* Perform initialization at beginning and after module has been re-loaded */
 		int rd, i;
@@ -844,7 +868,8 @@ int main(int argc, char **argv) {
 		ioctl(fileDesc, EVIOCGNAME(sizeof(name)), name);
 		printf("Input device name: \"%s\"\n", name);
 
-		XSetErrorHandler(invalidWindowHandler);
+		//TODO activate again?
+		//XSetErrorHandler(invalidWindowHandler);
 
 
 		int opcode, event, error;
@@ -875,10 +900,10 @@ int main(int argc, char **argv) {
 			exit(1);
 		}
 
-		/* Which version of XI2? We support 2.0 */
-		major = 2; minor = 0;
+		/* Which version of XI2? We support 2.1 */
+		major = 2; minor = 1;
 		if (XIQueryVersion(display, &major, &minor) == BadRequest) {
-			printf("XI2 not available. Server supports %d.%d\n", major, minor);
+			printf("XI 2.1 not available. Server supports %d.%d\n", major, minor);
 			XCloseDisplay(display);
 			exit(1);
 		}
@@ -921,166 +946,188 @@ int main(int argc, char **argv) {
 
 		/* Receive device property change events */
 		XIEventMask device_mask2;
-		unsigned char mask_data2[2] = { 0, 0 };
+		unsigned char mask_data2[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
 		device_mask2.deviceid = deviceID;
 		device_mask2.mask_len = sizeof(mask_data2);
 		device_mask2.mask = mask_data2;
 		XISetMask(device_mask2.mask, XI_PropertyEvent);
+		XISetMask(device_mask2.mask, XI_ButtonPress);
+		//XISetMask(device_mask2.mask, XI_TouchBegin);
+		//XISetMask(device_mask2.mask, XI_TouchUpdate);
+		//XISetMask(device_mask2.mask, XI_TouchEnd);
 		XISelectEvents(display, root, &device_mask2, 1);
 
 		/* Recieve events when screen size changes */
 		XRRSelectInput(display, root, RRScreenChangeNotifyMask);
 
+
+		/* Receive touch events */
+
+
+
 		/* Needed for XTest to work correctly */
 		XTestGrabControl(display, True);
-		active = 0;
-		checkRunningWindows();
-		Window w = getCurrentWindow();
-		activateAtRelease = 0;
-		if(debugMode) printf("Current Window: %i\n", (int) w);
 
-		if(isWindowBlacklisted(w)) {
-			/* Current is blacklisted, call enterBlacklistedWindow */
-			enterBlacklistedWindow();
-		} else {
-			/* Current is not blacklisted, so we might have left a blacklisted window */
-			activate();
-		}
 
 		/* Needed for some reason to receive events */
-		XGrabPointer(display, root, False, 0, GrabModeAsync, GrabModeAsync,
+/*		XGrabPointer(display, root, False, 0, GrabModeAsync, GrabModeAsync,
 				None, None, CurrentTime);
-		XUngrabPointer(display, CurrentTime);
+		XUngrabPointer(display, CurrentTime);*/
 
-		/* If it is not already running, create thread that listens to XInput events */
-		if (threadReturn != 0)
-			threadReturn = pthread_create(&xLoopThread, NULL,
-					xLoopThreadFunction, NULL);
+		grab(display, deviceID);		
+
+
+		///* If it is not already running, create thread that listens to XInput events */
+		//if (threadReturn != 0)
+		//	threadReturn = pthread_create(&xLoopThread, NULL,
+		//			xLoopThreadFunction, NULL);
 
 		printf("Reading input from device ... (interrupt to exit)\n");
 
-		/* If set, twofing assumes that the device uses the multitouch slot protocol.
-		   Set to 1 the first time a ABS_MT_SLOT event occurs. Set to 0 the first time a MT_SYNC event occurs.*/
-		int useSlots = 1;
 		int currentSlot = 0;
-		/* If we don't use the slot protocol, we collect all data of one finger into tempFingerInfo and set
+
+		/* If we use the legacy protocol, we collect all data of one finger into tempFingerInfo and set
 		   it to the correct slot once MT_SYNC occurs. */
-		FingerInfo tempFingerInfo = { -1, -1, -1, -1 };
+		FingerInfo tempFingerInfo = { .rawX=0, .rawY=0, .rawZ=0, .id = -1, .slotUsed = 0, .setThisTime = 0 };
 
-		/* Kernel device event loop */
 		while (1) {
-			rd = read(fileDesc, ev, sizeof(struct input_event) * 64);
-			if (rd < (int) sizeof(struct input_event)) {
-				printf("Data stream stopped\n");
-				break;
-			}
-			for (i = 0; i < rd / sizeof(struct input_event); i++) {
-				if (ev[i].type == EV_SYN) {
 
-					if (2 == ev[i].code) { // MT_Sync : Multitouch event end
-						/* Finger info for one finger collected in tempFingerInfo, so save it to fingerInfos. */
 
-						if(useSlots) {
-							/* This messsage indicates we don't use the slot protocol. So
-							   set useSlots and ignore data for now. */
-							useSlots = 0;
-							currentSlot = -1;
-							if(debugMode) printf("Switch to non-slot protocol.\n");
-						} else {
+			FD_SET(fileDesc, &fileDescSet);
+			FD_SET(eventQueueDesc, &fileDescSet);
 
-							/* Look for slot to put the data into by looking at the tracking ids */
-							int index = -1;
-							int i;
-							for(i = 0; i < 2; i++) {
-								if(fingerInfos[i].id == tempFingerInfo.id) {
-									index = i;
-									break;
-								}
-							}
-							
-							if(index == -1) {
+			select(MAX(fileDesc, eventQueueDesc) + 1, &fileDescSet, NULL, NULL, getEasingStepTimeVal());
+			
+			checkEasingStep();
+
+			if(FD_ISSET(fileDesc, &fileDescSet))
+			{
+
+
+				rd = read(fileDesc, ev, sizeof(struct input_event) * 64);
+				if (rd < (int) sizeof(struct input_event)) {
+					printf("Data stream stopped\n");
+					break;
+				}
+				for (i = 0; i < rd / sizeof(struct input_event); i++) {
+
+					if (ev[i].type == EV_SYN) {
+						if (0 == ev[i].code) { // Ev_Sync event end
+							/* All finger data received, so process now. */
+
+							if(useLegacyProtocol) {
+								/* Clear slots not set this time */
+								int i;
 								for(i = 0; i < 2; i++) {
-									if(fingerInfos[i].id == -1) {
-										/* "Empty" slot, so we can add it. */
-										index = i;
-										fingerInfos[i].id = tempFingerInfo.id;
-										break;
+									if(fingerInfos[i].setThisTime) {
+										fingerInfos[i].setThisTime = 0;
+									} else {
+										/* Clear slot */
+										fingerInfos[i].slotUsed = 0;
+									}
+								}
+								tempFingerInfo.slotUsed = 0;
+							}
+
+							processFingers();
+
+						} else if (2 == ev[i].code) { // MT_Sync : Multitouch event end
+
+							if(!useLegacyProtocol) {
+
+								/* This messsage indicates we use legacy protocol, so switch */
+								useLegacyProtocol = 1;
+								currentSlot = -1;
+								if(debugMode) printf("Switch to legacy protocol.\n");
+							} else {
+								if(tempFingerInfo.slotUsed) {
+									/* Finger info for one finger collected in tempFingerInfo, so save it to fingerInfos. */
+
+									/* Look for slot to put the data into by looking at the tracking ids */
+									int index = -1;
+									int i;
+									for(i = 0; i < 2; i++) {
+										if(fingerInfos[i].slotUsed && fingerInfos[i].id == tempFingerInfo.id) {
+											index = i;
+											break;
+										}
+									}
+							
+									if(index == -1) {
+										for(i = 0; i < 2; i++) {
+											if(!fingerInfos[i].slotUsed) {
+												/* "Empty" slot, so we can add it. */
+												index = i;
+												fingerInfos[i].id = tempFingerInfo.id;
+												fingerInfos[i].slotUsed = 1;
+												break;
+											}
+										}
+									}
+
+									if(index != -1) {
+										/* Copy temporary data to slot */
+										fingerInfos[index].setThisTime = 1;
+										fingerInfos[index].rawX = tempFingerInfo.rawX;
+										fingerInfos[index].rawY = tempFingerInfo.rawY;
+										fingerInfos[index].rawZ = tempFingerInfo.rawZ;
 									}
 								}
 							}
-
-							if(index != -1) {
-								fingerInfos[index].setThisTime = 1;
-								fingerInfos[index].rawX = tempFingerInfo.rawX;
-								fingerInfos[index].rawY = tempFingerInfo.rawY;
-							}
-						}
-					} else if (0 == ev[i].code) { // Ev_Sync event end
-
-						if(!useSlots) {
-							/* Clear slots not set this time */
-							int i;
-							for(i = 0; i < 2; i++) {
-								if(fingerInfos[i].setThisTime) {
-									fingerInfos[i].setThisTime = 0;
-								} else {
-									/* Clear slot */
-									fingerInfos[i].id = -1;
-								}
-							}
 						}
 
-
-						/* All finger data received, so process now. */
-						processFingers();
-
-						if(!useSlots) {
-							/* If we don't use the slot protocol, reset for next set of events */
-							tempFingerInfo.id = -1;
-						}
-					}
-				} else if (ev[i].type == EV_MSC && (ev[i].code == MSC_RAW
-						|| ev[i].code == MSC_SCAN)) {
-
-				} else if (ev[i].code == 47) {
-					/* Slot event */
-					if(!useSlots) {
-						useSlots = 1;
-						if(debugMode) printf("Switch to slots protocol.\n");
-					}
-					currentSlot = ev[i].value;
-					if(currentSlot < 0 || currentSlot > 1) currentSlot = -1;
-				} else {
-					/* Set finger info values for current finger */
-					if (ev[i].code == 57) {
-						/* ABS_MT_TRACKING_ID */
-						if(useSlots) {
+					} else if (ev[i].type == EV_MSC && (ev[i].code == MSC_RAW
+							|| ev[i].code == MSC_SCAN)) {
+					} else if (ev[i].code == 47) {
+						currentSlot = ev[i].value;
+						if(currentSlot < 0 || currentSlot > 1) currentSlot = -1;
+					} else {
+						/* Set finger info values for current finger */
+						if (ev[i].code == 57) {
+							/* ABS_MT_TRACKING_ID */
 							if(currentSlot != -1) {
-								fingerInfos[currentSlot].id = ev[i].value;
+								if(ev[i].value == -1)
+								{
+									fingerInfos[currentSlot].slotUsed = 0;
+								}
+								else
+								{
+									fingerInfos[currentSlot].id = ev[i].value;
+									fingerInfos[currentSlot].slotUsed = 1;
+								}
+							} else if(useLegacyProtocol) {
+								tempFingerInfo.id = ev[i].value;
+								tempFingerInfo.slotUsed = 1;
 							}
-						} else {
-							tempFingerInfo.id = ev[i].value;
-						}
-					};
-					if (ev[i].code == 53) {
-						if(useSlots) {
+						};
+						if (ev[i].code == 53) {
 							if(currentSlot != -1) {
 								fingerInfos[currentSlot].rawX = ev[i].value;
+							} else if(useLegacyProtocol) {
+								tempFingerInfo.rawX = ev[i].value;
 							}
-						} else {
-							tempFingerInfo.rawX = ev[i].value;
-						}
-					};
-					if (ev[i].code == 54) {
-						if(useSlots) {
+						};
+						if (ev[i].code == 54) {
 							if(currentSlot != -1) {
 								fingerInfos[currentSlot].rawY = ev[i].value;
+							} else if(useLegacyProtocol) {
+								tempFingerInfo.rawY = ev[i].value;
 							}
-						} else {
-							tempFingerInfo.rawY = ev[i].value;
-						}
-					};
+						};
+						if (ev[i].code == 58) {
+							if(currentSlot != -1) {
+								fingerInfos[currentSlot].rawZ = ev[i].value;
+							} else if(useLegacyProtocol) {
+								tempFingerInfo.rawZ = ev[i].value;
+							}
+						};
+					}
 				}
+
+			}
+
+			if(FD_ISSET(eventQueueDesc, &fileDescSet)) {
+				handleXEvent();
 			}
 		}
 
@@ -1088,10 +1135,8 @@ int main(int argc, char **argv) {
 		close(fileDesc);
 
 		/* Clean up */
-		if (active) {
-			ungrab(display, deviceID);
-		}
 		releaseButton();
+		ungrab(display, deviceID);
 
 		/* Wait until device file is there again */
 		while ((fileDesc = open(devname, O_RDONLY)) < 0) {
