@@ -1,5 +1,5 @@
 /*
- Copyright (C) 2010, Philipp Merkel <linux@philmerk.de>
+ Copyright (C) 2010, 2011 Philipp Merkel <linux@philmerk.de>
 
  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
@@ -17,6 +17,8 @@
 #ifndef TWOFINGEMU_H_
 #define TWOFINGEMU_H_
 
+#define MAX(a, b) (a > b ? a : b) 
+
 typedef struct FingerInfo FingerInfo;
 typedef struct Action Action;
 typedef struct Profile Profile;
@@ -26,7 +28,8 @@ struct FingerInfo {
 	int y;
 	int rawX;
 	int rawY;
-	int setThisTime;
+	int rawZ;
+	int slotUsed;
 	int id; /* Tracking ID */
 };
 
@@ -92,27 +95,23 @@ struct Profile {
 int inDebugMode();
 
 void *xLoopThreadFunction(void *arg);
-void windowMapped(Window w);
 int isWindowBlacklisted(Window w);
-void enterBlacklistedWindow();
-void leaveWindow();
 
 Window getCurrentWindow();
 char* getWindowClass(Window);
 Window getLastChildWindow(Window);
 
+Window getActiveWindow();
+
 void processFingers();
 
 void pressButton();
 void releaseButton();
 int isButtonDown();
 
-void movePointer(int, int);
+void movePointer(int, int, int);
 void executeAction(Action* action, int what);
 
-void deactivate();
-void activate();
-
 void ungrab(Display *display,int deviceid);
 void grab(Display *display,int deviceid);
 
@@ -121,7 +120,10 @@ int invalidWindowHandler(Display *dsp,XErrorEvent *err);
 void readCalibrationData();
 void startContinuation();
 
-Time getCurrentTime();
+typedef struct timeval TimeVal;
 
+TimeVal getCurrentTime();
+
+int timeDiff(TimeVal start, TimeVal end);
 
 #endif /* TWOFINGEMU_H_ */
